#!/bin/sh

print_perl_include_paths() {
    perl -e 'foreach (@INC) { print "\t$_\n"; }'
}

APPNAME=rebot3.pl

echo "Preparing to install $APPNAME. First let's do some checks."
echo

# perl executable check
echo -n "Checking for perl..."
perl -e 'print " OK\n"' 2> /dev/null
if [ $? -ne 0 ]; then
    echo " not found"
    cat <<EOF
There was error running perl. Make sure perl is installed on your system and
the directory containing your perl executable is in your PATH environment
variable.
EOF
    exit 1
fi

# perl version check
echo -n "Checking for correct perl version..."
perl -e 'require 5.004; print " OK (version $])\n"' 2> /dev/null
if [ $? -ne 0 ]; then
    perl -e 'print " incorrect (version $])\n"'
    cat <<EOF
$APPNAME requires at least perl version 5.004. Please update to a supported
version.
EOF
    exit 2
fi

# CDDB module check
echo -n "Checking for module CDDB..."
perl -MCDDB -e 'print " OK\n"' 2> /dev/null
if [ $? -ne 0 ]; then
    echo " not installed"
    cat <<EOF
The perl module CDDB is not installed correctly. Please install it and make
sure it is available in one of perl's include paths. You can get this module
from www.cpan.org. Or use search.cpan.org to let CPAN's search engine make
the dirty work for you.
Btw, perl's include paths are:
EOF
    print_perl_include_paths
    exit 3
fi

# MP3::ID3v1Tag module check
echo -n "Checking for module MP3::ID3v1Tag..."
perl -MMP3::ID3v1Tag -e 'print " OK\n"' 2> /dev/null
if [ $? -ne 0 ]; then
    echo " not installed"
    cat <<EOF
The perl module MP3v1Tag is not installed correctly. Please install it and
make sure it is available in one of perl's include paths. You can get this
module from www.cpan.org. Or use search.cpan.org to let CPAN's search engine
make the dirty work for you.
Btw, perl's include paths are:
EOF
    print_perl_include_paths
    exit 4
fi

echo
echo "Looks good. Proceeding with install."
echo

# default install paths
BIN_PATH=/usr/local/bin
MAN_PATH=/usr/local/man

# bin path prompt
echo "Enter the path where $APPNAME should be installed."
echo -n "[$BIN_PATH]: "
read bin_path
: ${bin_path:=$BIN_PATH}

# man path prompt
echo "Enter the path where $APPNAME's man page should be installed."
echo -n "[$MAN_PATH]: "
read man_path
: ${man_path:=$MAN_PATH}

# translate leading tilde to users home dir
bin_path=$(perl -e '$p = $ARGV[0]; $h = ($ENV{HOME} || $ENV{LOGDIR}); unless (defined $h) { print $p; exit; } $p =~ s/^~/$h/; print $p;' "$bin_path")
man_path=$(perl -e '$p = $ARGV[0]; $h = ($ENV{HOME} || $ENV{LOGDIR}); unless (defined $h) { print $p; exit; } $p =~ s/^~/$h/; print $p;' "$man_path")

# check for source files
for n in $APPNAME $APPNAME.1; do
    if ! [ -r $n ]; then
        echo "$n not found in current directory. Aborting installation..."
        exit 5
    fi
done

# the installation itself
error=0
install -d -m 0755 "$bin_path";                 let "error = error + $?"
install    -m 0755 $APPNAME "$bin_path";        let "error = error + $?"
install -d -m 0755 "$man_path/man1";            let "error = error + $?"
install    -m 0644 $APPNAME.1 "$man_path/man1"; let "error = error + $?"

# bye
echo
if [ $error -ne 0 ]; then
    cat <<EOF
Oops. Some error occurred while installing $APPNAME. Please investigate the
problem and run `basename $0` again.
EOF
    exit 6
else
    cat <<EOF
Congratulations. You just installed $APPNAME successfully.
Please send any comments or suggestions to panos@bigfoot.de.
Have fun!
EOF
fi
